import logging
import os
import pathlib
import uuid
from typing import Dict, Optional, List, Union

import pandas as pd
from quart import Quart, request, jsonify
from quart.helpers import stream_with_context
from pydantic import BaseModel, ValidationError

from autorag.deploy.base import BaseRunner
from autorag.nodes.generator.base import BaseGenerator
from autorag.utils import fetch_contents

logger = logging.getLogger("AutoRAG")

deploy_dir = pathlib.Path(__file__).parent
root_dir = pathlib.Path(__file__).parent.parent

VERSION_PATH = os.path.join(root_dir, "VERSION")


class QueryRequest(BaseModel):
	query: str
	result_column: Optional[str] = "generated_texts"


class RetrievedPassage(BaseModel):
	content: str
	doc_id: str
	filepath: Optional[str] = None
	file_page: Optional[int] = None
	start_idx: Optional[int] = None
	end_idx: Optional[int] = None


class RunResponse(BaseModel):
	result: Union[str, List[str]]
	retrieved_passage: List[RetrievedPassage]


class VersionResponse(BaseModel):
	version: str


empty_retrieved_passage = RetrievedPassage(
	content="", doc_id="", filepath=None, file_page=None, start_idx=None, end_idx=None
)


class ApiRunner(BaseRunner):
	def __init__(self, config: Dict, project_dir: Optional[str] = None):
		super().__init__(config, project_dir)
		self.app = Quart(__name__)

		data_dir = os.path.join(project_dir, "data")
		self.corpus_df = pd.read_parquet(
			os.path.join(data_dir, "corpus.parquet"), engine="pyarrow"
		)
		self.__add_api_route()

	def __add_api_route(self):
		@self.app.route("/v1/run", methods=["POST"])
		async def run_query():
			try:
				data = await request.get_json()
				data = QueryRequest(**data)
			except ValidationError as e:
				return jsonify(e.errors()), 400

			previous_result = pd.DataFrame(
				{
					"qid": str(uuid.uuid4()),
					"query": [data.query],
					"retrieval_gt": [[]],
					"generation_gt": [""],
				}
			)  # pseudo qa data for execution
			for module_instance, module_param in zip(
				self.module_instances, self.module_params
			):
				new_result = module_instance.pure(
					previous_result=previous_result, **module_param
				)
				duplicated_columns = previous_result.columns.intersection(
					new_result.columns
				)
				drop_previous_result = previous_result.drop(columns=duplicated_columns)
				previous_result = pd.concat([drop_previous_result, new_result], axis=1)

			# Simulate processing the query
			generated_text = previous_result[data.result_column].tolist()[0]
			retrieved_passage = self.extract_retrieve_passage(previous_result)

			response = RunResponse(
				result=generated_text, retrieved_passage=retrieved_passage
			)

			return jsonify(response.model_dump()), 200

		@self.app.route("/v1/stream", methods=["POST"])
		async def stream_query():
			try:
				data = await request.get_json()
				data = QueryRequest(**data)
			except ValidationError as e:
				return jsonify(e.errors()), 400

			@stream_with_context
			async def generate():
				previous_result = pd.DataFrame(
					{
						"qid": str(uuid.uuid4()),
						"query": [data.query],
						"retrieval_gt": [[]],
						"generation_gt": [""],
					}
				)  # pseudo qa data for execution

				for module_instance, module_param in zip(
					self.module_instances, self.module_params
				):
					if not isinstance(module_instance, BaseGenerator):
						new_result = module_instance.pure(
							previous_result=previous_result, **module_param
						)
						duplicated_columns = previous_result.columns.intersection(
							new_result.columns
						)
						drop_previous_result = previous_result.drop(
							columns=duplicated_columns
						)
						previous_result = pd.concat(
							[drop_previous_result, new_result], axis=1
						)
					else:
						retrieved_passages = self.extract_retrieve_passage(
							previous_result
						)
						response = RunResponse(
							result="", retrieved_passage=retrieved_passages
						)
						yield response.model_dump_json().encode("utf-8")
						# Start streaming of the result
						assert len(previous_result) == 1
						prompt: str = previous_result["prompts"].tolist()[0]
						async for delta in module_instance.stream(
							prompt=prompt, **module_param
						):
							response = RunResponse(
								result=delta,
								retrieved_passage=[empty_retrieved_passage],
							)
							yield response.model_dump_json().encode("utf-8")

			return generate(), 200, {"X-Something": "value"}

		@self.app.route("/version", methods=["GET"])
		def get_version():
			with open(VERSION_PATH, "r") as f:
				version = f.read().strip()
			response = VersionResponse(version=version)
			return jsonify(response.model_dump()), 200

	def run_api_server(self, host: str = "0.0.0.0", port: int = 8000, **kwargs):
		"""
		Run the pipeline as api server.
		You can send POST request to `http://host:port/run` with json body like below:

		.. Code:: json

		    {
		        "query": "your query",
		        "result_column": "generated_texts"
		    }

		And it returns json response like below:

		.. Code:: json

		    {
		        "answer": "your answer"
		    }

		:param host: The host of the api server.
		:param port: The port of the api server.
		:param kwargs: Other arguments for Flask app.run.
		"""
		logger.info(f"Run api server at {host}:{port}")
		self.app.run(host=host, port=port, **kwargs)

	def extract_retrieve_passage(self, df: pd.DataFrame) -> List[RetrievedPassage]:
		retrieved_ids: List[str] = df["retrieved_ids"].tolist()[0]
		contents = fetch_contents(self.corpus_df, [retrieved_ids])[0]
		if "path" in self.corpus_df.columns:
			paths = fetch_contents(self.corpus_df, [retrieved_ids], column_name="path")[
				0
			]
		else:
			paths = [None] * len(retrieved_ids)
		metadatas = fetch_contents(
			self.corpus_df, [retrieved_ids], column_name="metadata"
		)[0]
		if "start_end_idx" in self.corpus_df.columns:
			start_end_indices = fetch_contents(
				self.corpus_df, [retrieved_ids], column_name="start_end_idx"
			)[0]
		else:
			start_end_indices = [None] * len(retrieved_ids)
		return list(
			map(
				lambda content, doc_id, path, metadata, start_end_idx: RetrievedPassage(
					content=content,
					doc_id=doc_id,
					filepath=path,
					file_page=metadata.get("page", None),
					start_idx=start_end_idx[0] if start_end_idx else None,
					end_idx=start_end_idx[1] if start_end_idx else None,
				),
				contents,
				retrieved_ids,
				paths,
				metadatas,
				start_end_indices,
			)
		)



from pathlib import Path
from typing import Any, List, Tuple

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm

from autorag.nodes.passagereranker.base import BasePassageReranker

from transformers import AutoTokenizer
from optimum.intel.openvino import OVModelForSequenceClassification


from autorag.utils.util import (
	make_batch,
	sort_by_scores,
	flatten_apply,
	select_top_k,
	result_to_dataframe,
	pop_params,
)


class OpenVINOReranker(BasePassageReranker):
	def __init__(
		self,
		project_dir: str,
		model: str = "BAAI/bge-reranker-large",
		*args,
		**kwargs,
	):
		super().__init__(project_dir)

		try:
			from huggingface_hub import HfApi
		except ImportError as e:
			raise ValueError(
				"Could not import huggingface_hub python package. "
				"Please install it with: "
				"`pip install -U huggingface_hub`."
			) from e

		def require_model_export(
			model_id: str, revision: Any = None, subfolder: Any = None
		) -> bool:
			model_dir = Path(model_id)
			if subfolder is not None:
				model_dir = model_dir / subfolder
			if model_dir.is_dir():
				return (
					not (model_dir / "openvino_model.xml").exists()
					or not (model_dir / "openvino_model.bin").exists()
				)
			hf_api = HfApi()
			try:
				model_info = hf_api.model_info(model_id, revision=revision or "main")
				normalized_subfolder = (
					None if subfolder is None else Path(subfolder).as_posix()
				)
				model_files = [
					file.rfilename
					for file in model_info.siblings
					if normalized_subfolder is None
					or file.rfilename.startswith(normalized_subfolder)
				]
				ov_model_path = (
					"openvino_model.xml"
					if subfolder is None
					else f"{normalized_subfolder}/openvino_model.xml"
				)
				return (
					ov_model_path not in model_files
					or ov_model_path.replace(".xml", ".bin") not in model_files
				)
			except Exception:
				return True

		model_kwargs = pop_params(
			OVModelForSequenceClassification.from_pretrained, kwargs
		)

		if require_model_export(model):
			# use remote model
			self.model = OVModelForSequenceClassification.from_pretrained(
				model, export=True, **model_kwargs
			)
		else:
			# use local model
			self.model = OVModelForSequenceClassification.from_pretrained(
				model, **model_kwargs
			)

		self.tokenizer = AutoTokenizer.from_pretrained(model)

	def __del__(self):
		del self.model
		del self.tokenizer
		if torch.cuda.is_available():
			torch.cuda.empty_cache()
		super().__del__()

	@result_to_dataframe(["retrieved_contents", "retrieved_ids", "retrieve_scores"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, contents, _, ids = self.cast_to_run(previous_result)
		top_k = kwargs.get("top_k", 3)
		batch = kwargs.get("batch", 64)
		return self._pure(queries, contents, ids, top_k, batch)

	def _pure(
		self,
		queries: List[str],
		contents_list: List[List[str]],
		ids_list: List[List[str]],
		top_k: int,
		batch: int = 64,
	) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
		"""
		Rerank a list of contents based on their relevance to a query using MonoT5.

		:param queries: The list of queries to use for reranking
		:param contents_list: The list of lists of contents to rerank
		:param ids_list: The list of lists of ids retrieved from the initial ranking
		:param top_k: The number of passages to be retrieved

		:param batch: The number of queries to be processed in a batch
		:return: tuple of lists containing the reranked contents, ids, and scores
		"""
		# Retrieve the tokens used by the model to represent false and true predictions

		nested_list = [
			list(map(lambda x: [query, x], content_list))
			for query, content_list in zip(queries, contents_list)
		]

		rerank_scores = flatten_apply(
			openvino_run_model,
			nested_list,
			model=self.model,
			batch_size=batch,
			tokenizer=self.tokenizer,
		)

		df = pd.DataFrame(
			{
				"contents": contents_list,
				"ids": ids_list,
				"scores": rerank_scores,
			}
		)
		df[["contents", "ids", "scores"]] = df.apply(
			sort_by_scores, axis=1, result_type="expand"
		)
		results = select_top_k(df, ["contents", "ids", "scores"], top_k)

		return (
			results["contents"].tolist(),
			results["ids"].tolist(),
			results["scores"].tolist(),
		)


def openvino_run_model(
	input_texts,
	model,
	batch_size: int,
	tokenizer,
):
	batch_input_texts = make_batch(input_texts, batch_size)
	results = []
	for batch_texts in tqdm(batch_input_texts):
		input_tensors = tokenizer(
			batch_texts,
			padding=True,
			truncation=True,
			return_tensors="pt",
		)

		outputs = model(**input_tensors, return_dict=True)
		if outputs[0].shape[1] > 1:
			scores = outputs[0][:, 1]
		else:
			scores = outputs[0].flatten()

		scores = list(map(float, (1 / (1 + np.exp(-np.array(scores))))))
		results.extend(scores)
	return results
